module treemod
    implicit none
    type :: person
        character(20) :: vname = "                    "  ! Vorname
        character(20) :: nname = "                    "  ! Nachname
        integer :: jahr = 0   ! Geburtsjahr
    end type

    type :: node 
        type(person) :: daten
        type(node), pointer :: left => null(), right => null()
    end type

    type :: tree
        type(node), pointer :: root => null()
    end type 
contains

recursive subroutine Ausgabe(Blatt)
    type(node),intent(in)::Blatt
    
    If (associated(Blatt%left)) then 
        call Ausgabe(Blatt%Left)
    end if 

    write(*,*)Blatt%daten%vname, Blatt%daten%nname, Blatt%daten%jahr

    If (associated(Blatt%right)) then 
        call Ausgabe(Blatt%right)
    end if 
end subroutine Ausgabe 
        
function newPerson(vname, nname, jahr)
    character(20), intent(in):: vname  ! Vorname
    character(20), intent(in) :: nname  ! Nachname
    integer, intent(in) :: jahr
    type(person) :: newPerson
    newPerson%vname = vname
    newPerson%nname = nname
    newPerson%jahr = jahr
end function

subroutine printPerson (P)
    type(person), intent(in) :: P
    write(*,"(A)",advance="no") trim(P%vname)//" "//trim(P%nname)//", "
    write(*,"(I4)") P%jahr
end subroutine



function newTree(rootPerson, left, right)
    type(tree), intent(in), optional :: left, right
    type(person), intent(in) :: rootPerson
    type(tree) :: newTree
    allocate(newTree%root)
    newTree%root%daten = rootPerson
    if ( present(left) ) then
        newTree%root%left => left%root
    end if
    if ( present(right) ) then
        newTree%root%right => right%root
    end if
end function newTree


subroutine invertTree(baum)
    type(tree),intent(in)::baum
    call invertRoot(baum%root)
end subroutine

recursive subroutine invertRoot(blatt)
    type(node),pointer::blatt
    call tradeParents(blatt)
    if(associated(blatt%right))then
        call invertRoot(blatt%right)
    end if
    if(associated(blatt%left))then
        call invertRoot(blatt%left)
    end if
end subroutine

subroutine tradeParents(blatt)
    type(node),intent(inout)::blatt
    type(node)::speicher
    speicher%right=>blatt%right
    blatt%right=>blatt%left
    blatt%left=>speicher%right
end subroutine

subroutine printTree(T)
    type(tree), intent(in) :: T
    call recPrintTree(T%root, 0)
contains
    recursive subroutine recPrintTree(N, depth)
        type(node), pointer, intent(in) :: N 
        integer, intent(in) :: depth
        integer :: i
        if ( associated(N) ) then
            call recPrintTree(N%left, depth+5)
            do i = 1, depth
                write(*,"(X)", advance="no")
            end do
            call printPerson(N%daten)
            call recPrintTree(N%right, depth+5)
        end if
    end subroutine recPrintTree
end subroutine printTree

function groesse(T) result(g)
    type(tree), intent(in) :: T
    integer :: g

    g = nodeGroesse(T%root)

    contains

    recursive function nodeGroesse(N) result(ng)
        type(node), pointer, intent(in) :: N
        integer :: ng
        
        if (.not. associated(N)) then
            ng = 0
        else
            ng = nodeGroesse(N%left) + nodeGroesse(N%right) + 1 
        end if
    end function nodeGroesse
end function

function hoehe(T) result(h)
    type(tree), intent(in) :: T
    integer :: h
    
    h = nodeHoehe(T%root)

    contains
    
    recursive function nodeHoehe(N) result(nh)
        type(node), pointer, intent(in) :: N
        integer :: nh
        
        if (.not. associated(N)) then
            nh = 0
        else
            nh = max(nodeHoehe(N%left), nodeHoehe(N%right)) + 1
        end if
    end function nodeHoehe
end function hoehe

subroutine deleteTree(T)
    type(tree), intent(inout) :: T
    call recDeleteTree(T%root)
contains
    recursive subroutine recDeleteTree(N)
        type(node), pointer, intent(inout) :: N 
        if ( associated(N) ) then
            call recDeleteTree(N%left)
            call recDeleteTree(N%right)
            deallocate(N)
        end if
    end subroutine recDeleteTree
end subroutine deleteTree

subroutine search(searched, tosearch, help)
    type(person), intent(in) :: searched
    type(tree), intent(in) :: tosearch
    type(node), intent(out), pointer :: help
    
    help => null()
    
    call assistant(searched, tosearch%root, help)
    
    
contains

    recursive subroutine assistant(searched, tosearch, help)
        type(person), intent(in) :: searched
        type(node), intent(in), pointer :: tosearch
        type(node), intent(out), pointer :: help => null()
    
        if ( equal(tosearch%daten, searched) ) then
            help => tosearch
        else
            if ( associated(tosearch%left) ) then
                call assistant(searched, tosearch%left, help)
            end if

            if ( .not. associated(help) .and. associated(tosearch%right) ) then
                call assistant(searched, tosearch%right, help)
            end if

        end if
    end subroutine assistant

    function equal(arg1, arg2) result(retval)
        type(person), intent(in) :: arg1, arg2
        logical :: retval
    
        retval = arg1%vname == arg2%vname .and. arg1%nname == arg2%nname .and. arg1%jahr == arg2%jahr
        
    end function equal
    
end subroutine search
end module treemod

program main
    use treemod
    implicit none
    type(tree) :: T 
    type(node), pointer :: holdSearchResult => null()
    character(20), dimension(5) :: vnames
    character(20), dimension(2) :: nnames
    vnames(1) = "Ulrich"
    vnames(2) = "Hertha" 
    vnames(3) = "Barbara"
    vnames(4) = "Florian"
    vnames(5) = "Bilbo"
    nnames(1) = "Schmidt"
    nnames(2) = "Lauche"

    T = newTree( newPerson(vnames(1),nnames(1),2111),                    &
    &            newTree( newPerson(vnames(2),nnames(1),2086),           &
    &                     newTree(newPerson(vnames(3),nnames(2),2049)),  &
    &                     newTree(newPerson(vnames(4),nnames(2),2046))), &
    &            newTree(newPerson(vnames(5),nnames(1),2091)))
    write(*,*) "Tree build"
    call printTree(T)
    write(*,*) "Groesse:",groesse(T)
    write(*,*) "Hoehe:", hoehe(T)
    write(*,*)
    write(*,*) "1. Suche (Erfolg erwartet):"
    call search(newPerson(vnames(2),nnames(1),2086),T,holdSearchResult)
    if (associated(holdSearchResult)) then
        write(*,*) "Person gefunden:", holdSearchResult%daten
    else
        write(*,*) "Person nicht gefunden."
    end if
    write(*,*) "2. Suche (Miserfolg erwartet):"
    call search(newPerson(vnames(2),nnames(1),-15),T,holdSearchResult)
    if (associated(holdSearchResult)) then
        write(*,*) "Person gefunden:", holdSearchResult%daten
    else
        write(*,*) "Person nicht gefunden."
    end if
    write(*,*) 
    write(*,*) "Invertierter Baum:"
    write(*,*)
    call invertTree(T)
    call printTree(T)
    call deleteTree(T)
    ! call printTree(T)
end program main